@@ -1,6 +1,6 @@
 /* etherinfo.c - Retrieve ethernet interface info via NETLINK
  *
- * Copyright (C) 2009-2010 Red Hat Inc.
+ * Copyright (C) 2009-2011 Red Hat Inc.
  *
  * David Sommerseth <davids@redhat.com>
  * Parts of this code is based on ideas and solutions in iproute2
@@ -21,15 +21,19 @@
 #include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <stdlib.h>
 #include <asm/types.h>
 #include <sys/socket.h>
 #include <netlink/route/rtnl.h>
 #include <assert.h>
 #include <errno.h>
+#include <pthread.h>
 #include "etherinfo_struct.h"
 #include "etherinfo.h"
 
+pthread_mutex_t nlc_counter_mtx = PTHREAD_MUTEX_INITIALIZER;
+
 /*
  *
  *   Internal functions for working with struct etherinfo
@@ -59,16 +63,14 @@
 void free_ipv6addresses(struct ipv6address *ptr) {
 	struct ipv6address *ipv6ptr = ptr;
 
-	if( !ptr ) {
-		return;
-	}
-
 	while( ipv6ptr ) {
 		struct ipv6address *tmp = ipv6ptr->next;
 
 		if( ipv6ptr->address ) {
 			free(ipv6ptr->address);
+			ipv6ptr->address = NULL;
 		}
+		memset(ipv6ptr, 0, sizeof(struct ipv6address));
 		free(ipv6ptr);
 		ipv6ptr = tmp;
 	}
@@ -147,7 +149,7 @@ static void callback_nl_link(struct nl_object *obj, void *arg)
 	unsigned char *binaddr;
 	char hwaddr[130], *ptr;
 
-	if( (ethi == NULL) || (ethi->hwaddress != NULL) ) {
+	if( (ethi == NULL) || (ethi->hwaddress != NULL) || (addr == NULL) ) {
 		return;
 	}
 
@@ -276,15 +278,25 @@ void dump_etherinfo(FILE *fp, struct etherinfo *ptr)
  *
  * @return Returns 1 on success, otherwise 0.
  */
-int get_etherinfo(struct etherinfo *ethinf, struct nl_handle *nlc, nlQuery query)
+int get_etherinfo(struct etherinfo_obj_data *data, nlQuery query)
 {
 	struct nl_cache *link_cache;
 	struct nl_cache *addr_cache;
 	struct rtnl_addr *addr;
 	struct rtnl_link *link;
+	struct etherinfo *ethinf = NULL;
 	int ret = 0;
 
-	if( !ethinf || !nlc ) {
+	if( !data || !data->ethinfo ) {
+		return 0;
+	}
+	ethinf = data->ethinfo;
+
+	/* Open a NETLINK connection on-the-fly */
+	if( !open_netlink(data) ) {
+		PyErr_Format(PyExc_RuntimeError,
+			     "Could not open a NETLINK connection for %s",
+			     ethinf->device);
 		return 0;
 	}
 
@@ -293,7 +305,7 @@ int get_etherinfo(struct etherinfo *ethinf, struct nl_handle *nlc, nlQuery query
 	 * interface index if we have that
 	 */
 	if( ethinf->index < 0 ) {
-		link_cache = rtnl_link_alloc_cache(nlc);
+		link_cache = rtnl_link_alloc_cache(*data->nlc);
 		ethinf->index = rtnl_link_name2i(link_cache, ethinf->device);
 		if( ethinf->index < 0 ) {
 			return 0;
@@ -305,7 +317,7 @@ int get_etherinfo(struct etherinfo *ethinf, struct nl_handle *nlc, nlQuery query
 	switch( query ) {
 	case NLQRY_LINK:
 		/* Extract MAC/hardware address of the interface */
-		link_cache = rtnl_link_alloc_cache(nlc);
+		link_cache = rtnl_link_alloc_cache(*data->nlc);
 		link = rtnl_link_alloc();
 		rtnl_link_set_ifindex(link, ethinf->index);
 		nl_cache_foreach_filter(link_cache, (struct nl_object *)link, callback_nl_link, ethinf);
@@ -316,9 +328,17 @@ int get_etherinfo(struct etherinfo *ethinf, struct nl_handle *nlc, nlQuery query
 
 	case NLQRY_ADDR:
 		/* Extract IP address information */
-		addr_cache = rtnl_addr_alloc_cache(nlc);
+		addr_cache = rtnl_addr_alloc_cache(*data->nlc);
 		addr = rtnl_addr_alloc();
 		rtnl_addr_set_ifindex(addr, ethinf->index);
+
+                /* Make sure we don't have any old IPv6 addresses saved */
+                if( ethinf->ipv6_addresses ) {
+                        free_ipv6addresses(ethinf->ipv6_addresses);
+                        ethinf->ipv6_addresses = NULL;
+                }
+
+                /* Retrieve all address information */
 		nl_cache_foreach_filter(addr_cache, (struct nl_object *)addr, callback_nl_address, ethinf);
 		rtnl_addr_put(addr);
 		nl_cache_free(addr_cache);
@@ -331,3 +351,82 @@ int get_etherinfo(struct etherinfo *ethinf, struct nl_handle *nlc, nlQuery query
 	return ret;
 }
 
+
+/**
+ * Connects to the NETLINK interface.  This will be called
+ * for each etherinfo object being generated, and it will
+ * keep a separate file descriptor open for each object
+ *
+ * @param data etherinfo_obj_data structure
+ *
+ * @return Returns 1 on success, otherwise 0.
+ */
+int open_netlink(struct etherinfo_obj_data *data)
+{
+	if( !data ) {
+		return 0;
+	}
+
+	/* Reuse already established NETLINK connection, if a connection exists */
+	if( *data->nlc ) {
+		/* If this object has not used NETLINK earlier, tag it as a user */
+		if( !data->nlc_active ) {
+			pthread_mutex_lock(&nlc_counter_mtx);
+			(*data->nlc_users)++;
+			pthread_mutex_unlock(&nlc_counter_mtx);
+		}
+		data->nlc_active = 1;
+		return 1;
+	}
+
+	/* No earlier connections exists, establish a new one */
+	*data->nlc = nl_handle_alloc();
+	nl_connect(*data->nlc, NETLINK_ROUTE);
+	if( (*data->nlc != NULL) ) {
+		/* Force O_CLOEXEC flag on the NETLINK socket */
+		if( fcntl(nl_socket_get_fd(*data->nlc), F_SETFD, FD_CLOEXEC) == -1 ) {
+			fprintf(stderr,
+				"**WARNING** Failed to set O_CLOEXEC on NETLINK socket: %s\n",
+				strerror(errno));
+		}
+
+		/* Tag this object as an active user */
+		pthread_mutex_lock(&nlc_counter_mtx);
+		(*data->nlc_users)++;
+		pthread_mutex_unlock(&nlc_counter_mtx);
+		data->nlc_active = 1;
+		return 1;
+	} else {
+		return 0;
+	}
+}
+
+
+/**
+ * Closes the NETLINK connection.  This should be called automatically whenever
+ * the corresponding etherinfo object is deleted.
+ *
+ * @param ptr  Pointer to the pointer of struct nl_handle, which contains the NETLINK connection
+ */
+void close_netlink(struct etherinfo_obj_data *data)
+{
+	if( !data || !(*data->nlc) ) {
+		return;
+	}
+
+	/* Untag this object as a NETLINK user */
+	data->nlc_active = 0;
+	pthread_mutex_lock(&nlc_counter_mtx);
+	(*data->nlc_users)--;
+	pthread_mutex_unlock(&nlc_counter_mtx);
+
+	/* Don't close the connection if there are more users */
+	if( *data->nlc_users > 0) {
+		return;
+	}
+
+	/* Close NETLINK connection */
+	nl_close(*data->nlc);
+	nl_handle_destroy(*data->nlc);
+	*data->nlc = NULL;
+}
