@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2010 Red Hat Inc.
+ * Copyright (C) 2008-2011 Red Hat Inc.
  *
  * Arnaldo Carvalho de Melo <acme@redhat.com>
  * David Sommerseth <davids@redhat.com>
@@ -32,6 +32,7 @@
 #include "etherinfo.h"
 
 static struct nl_handle *nlconnection = NULL;
+unsigned int nlconnection_users = 0;  /* How many NETLINK users are active? */
 extern PyTypeObject ethtool_etherinfoType;
 extern PyTypeObject ethtool_etherinfoIPv6Type;
 
@@ -93,8 +94,11 @@ static PyObject *get_active_devices(PyObject *self __unused, PyObject *args __un
 	ifr = ifc.ifc_req;
 	for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
 		if (!(ioctl(skfd, SIOCGIFFLAGS, ifr) < 0))
-			if (ifr->ifr_flags & IFF_UP)
-				PyList_Append(list, PyString_FromString(ifr->ifr_name));
+			if (ifr->ifr_flags & IFF_UP) {
+				PyObject *str = PyString_FromString(ifr->ifr_name);
+				PyList_Append(list, str);
+				Py_DECREF(str);
+			}
 			ifr++;
 	}
 
@@ -118,6 +122,7 @@ static PyObject *get_devices(PyObject *self __unused, PyObject *args __unused)
 	/* skip over first two lines */
 	ret = fgets(buffer, 256, fd); ret = fgets(buffer, 256, fd);
 	while (!feof(fd)) {
+		PyObject *str;
 		char *name = buffer;
 		char *end = buffer;
 
@@ -129,7 +134,10 @@ static PyObject *get_devices(PyObject *self __unused, PyObject *args __unused)
 		*end = 0; /* terminate where colon was */
 		while (*name == ' ')
 			name++; /* skip over leading whitespace if any */
-		PyList_Append(list, PyString_FromString(name));
+
+		str = PyString_FromString(name);
+		PyList_Append(list, str);
+		Py_DECREF(str);
 	}
 	fclose(fd);
 	return list;
@@ -307,7 +315,8 @@ static PyObject *get_interfaces_info(PyObject *self __unused, PyObject *args) {
 		 */
 		objdata->ethinfo->device = strdup(fetch_devs[i]);
 		objdata->ethinfo->index = -1;
-		objdata->nlc = nlconnection; /* Global variable */
+		objdata->nlc = &nlconnection;
+		objdata->nlc_users = &nlconnection_users;
 
 		/* Instantiate a new etherinfo object with the device information */
 		ethinf_py = PyCObject_FromVoidPtr(objdata, NULL);
@@ -320,7 +329,9 @@ static PyObject *get_interfaces_info(PyObject *self __unused, PyObject *args) {
 			PyObject *dev = PyObject_CallObject((PyObject *)&ethtool_etherinfoType, args);
 			if( dev ) {
 				PyList_Append(devlist, dev);
+				Py_DECREF(dev);
 			}
+			Py_DECREF(args);
 		}
 	}
 	if( fetch_devs_len > 0 ) {
@@ -970,52 +981,6 @@ static struct PyMethodDef PyEthModuleMethods[] = {
 };
 
 
-/**
- * Connects to the NETLINK interface.  This should only be
- * called once as part of the main ethtool module init.
- *
- * @param nlc Structure which keeps the NETLINK connection handle (struct nl_handle)
- *
- * @return Returns 1 on success, otherwise 0.
- */
-int open_netlink(struct nl_handle **nlc)
-{
-	if( *nlc ) {
-		return 0;
-	}
-
-	*nlc = nl_handle_alloc();
-	nl_connect(*nlc, NETLINK_ROUTE);
-	return (*nlc != NULL);
-}
-
-
-/**
- * Closes the NETLINK connection.  This should be called automatically whenever
- * the ethtool module is unloaded from Python.
- *
- * @param ptr  Pointer to the pointer of struct nl_handle, which contains the NETLINK connection
- */
-void close_netlink(void **ptr)
-{
-	struct nl_handle *nlc;
-
-	if( !ptr && !*ptr ) {
-		return;
-	}
-
-	nlc = (struct nl_handle *) *ptr;
-	if( !nlc ) {
-		return;
-	}
-
-	/* Close NETLINK connection */
-	nl_close(nlc);
-	nl_handle_destroy(nlc);
-	*ptr = NULL; /* reset the pointers pointer address */
-}
-
-
 PyMODINIT_FUNC initethtool(void)
 {
 	PyObject *m;
@@ -1033,12 +998,6 @@ PyMODINIT_FUNC initethtool(void)
 	Py_INCREF(&ethtool_etherinfoIPv6Type);
 	PyModule_AddObject(m, "etherinfo_ipv6addr", (PyObject *)&ethtool_etherinfoIPv6Type);
 
-	// Prepare an internal netlink connection object
-	if( open_netlink(&nlconnection) ) {
-		PyModule_AddObject(m, "__nlconnection",
-				   PyCObject_FromVoidPtr(&nlconnection, close_netlink));
-	}
-
 	// Setup constants
 	PyModule_AddIntConstant(m, "IFF_UP", IFF_UP);			/* Interface is up. */
 	PyModule_AddIntConstant(m, "IFF_BROADCAST", IFF_BROADCAST);	/* Broadcast address valid. */
@@ -1058,5 +1017,6 @@ PyMODINIT_FUNC initethtool(void)
 	PyModule_AddIntConstant(m, "IFF_DYNAMIC", IFF_DYNAMIC);		/* Dialup device with changing addresses.  */
 	PyModule_AddIntConstant(m, "AF_INET", AF_INET);                 /* IPv4 interface */
 	PyModule_AddIntConstant(m, "AF_INET6", AF_INET6);               /* IPv6 interface */
+	PyModule_AddStringConstant(m, "version", "python-ethtool v" VERSION);
 }
 
